import numpy as np
import pytest
import torch as th


# Raise errors for numerical issues.
np.seterr(divide="raise", over="raise", invalid="raise")
th.set_default_dtype(th.float64)


@pytest.fixture(params=[False, True], ids=["numpy", "torch"])
def use_torch(request: pytest.FixtureRequest) -> bool:
    return request.param


# Include fixtures that may be shared across different packages.
pytest_plugins = [
   "gptools.util.testing",
]



import matplotlib
import matplotlib.axes
import matplotlib.collections
import matplotlib.lines
import numpy as np
from typing import Optional, Tuple


def plot_band(x: np.ndarray, ys: np.ndarray, *, p: float = 0.05, relative_alpha: float = 0.25,
              ax: Optional[matplotlib.axes.Axes] = None, **kwargs) \
        -> Tuple[matplotlib.lines.Line2D, matplotlib.collections.PolyCollection]:
    """
    Plot a credible band given posterior samples.

    Args:
        x: Coordinates of samples with shape :code:`(n,)`, where :code:`n` is the number of
            observations.
        ys: Samples with shape :code:`(m, n)`, where :code:`m` is the number of samples.
        p: Tail probability to exclude from the credible band such that the band spans the quantiles
            :code:`[p / 2, 1 - p / 2]`.
        relative_alpha: Opacity of the band relative to the median line.
        ax: Axes to plot into.
        **kwargs: Keyword arguments passed to :code:`ax.plot` for the median line.

    Returns:
        line: Median line.
        band: Credible band spanning the quantiles :code:`[p / 2, 1 - p / 2]`.
    """
    from matplotlib import pyplot as plt
    ax = ax or plt.gca()
    l, m, u = np.quantile(ys, [p / 2, 0.5, 1.0 - p / 2], axis=0)
    line, = ax.plot(x, m, **kwargs)
    alpha = kwargs.get("alpha", 1.0) * relative_alpha
    return line, ax.fill_between(x, l, u, color=line.get_color(), alpha=alpha)



import math
from typing import Optional
from .. import ArrayOrTensor, ArrayOrTensorDispatch, mutually_exclusive_kwargs, \
    OptionalArrayOrTensor
from .util import log2, log_prob_stdnorm, sqrt2


dispatch = ArrayOrTensorDispatch()


def _get_rfft_scale(cov_rfft: OptionalArrayOrTensor, cov: OptionalArrayOrTensor,
                    rfft_scale: OptionalArrayOrTensor, size: Optional[int]) -> ArrayOrTensor:
    num_given = sum([cov_rfft is not None, cov is not None, rfft_scale is not None])
    if num_given != 1:  # pragma: no cover
        raise ValueError("exactly one of `cov_rfft`, `cov`, or `rfft_scale` must be given")
    if rfft_scale is not None:
        return rfft_scale
    return evaluate_rfft_scale(cov_rfft=cov_rfft, cov=cov, size=size)


@mutually_exclusive_kwargs("cov", "cov_rfft")
def evaluate_rfft_scale(*, cov: OptionalArrayOrTensor = None,
                        cov_rfft: OptionalArrayOrTensor = None, size: Optional[int] = None) \
        -> ArrayOrTensor:
    """
    Evaluate the scale of Fourier coefficients.

    Args:
        cov_rfft: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., size // 2 + 1)`.
        cov: First row of the covariance matrix with shape :code:`(..., size)`.
        size: Size of the real signal. Necessary because the size cannot be inferred from
            :code:`rfft`.

    Returns:
        scale: Scale of Fourier coefficients with shape :code:`(..., size // 2 + 1)`.
    """
    if cov_rfft is None:
        *_, size = cov.shape
        cov_rfft = dispatch[cov].fft.rfft(cov).real
    scale: ArrayOrTensor = dispatch.sqrt(size * cov_rfft / 2)
    # Rescale for the real-only zero frequency term.
    scale[0] *= sqrt2
    if size % 2 == 0:
        # Rescale for the real-only Nyqvist frequency term.
        scale[..., -1] *= sqrt2
    return scale


def expand_rfft(rfft: ArrayOrTensor, n: int) -> ArrayOrTensor:
    """
    Convert truncated real Fourier coefficients to full Fourier coefficients.

    Args:
        rfft: Truncated real Fourier coefficients with shape :code:`(n // 2 + 1,)`.
        n: Number of samples.

    Returns:
        fft: Full Fourier coefficients with shape :code:`(n,)`.
    """
    nrfft = n // 2 + 1
    ncomplex = (n - 1) // 2
    fft = dispatch[rfft].empty(n, dtype=rfft.dtype)
    fft[:nrfft] = rfft
    fft[nrfft:] = dispatch.flip(rfft[1:1 + ncomplex]).conj()
    return fft


def unpack_rfft(z: ArrayOrTensor, size: int) -> ArrayOrTensor:
    """
    Unpack the Fourier coefficients of a real Fourier transform with :code:`size // 2 + 1` elements
    to a vector of :code:`size` elements.

    Args:
        z: Real Fourier transform coefficients.
        size: Size of the real signal. Necessary because the size cannot be inferred from
            :code:`rfft`.

    Returns:
        z: Unpacked vector of :code:`size` elements comprising the :code:`size // 2 + 1` real parts
            of the zero frequency term, complex terms, and Nyqvist frequency term (for even
            :code:`size`). The subsequent :code:`(size - 1) // 2` elements are the imaginary parts
            of complex coefficients.
    """
    ncomplex = (size - 1) // 2
    parts = [z.real, z.imag[..., 1: ncomplex + 1]]
    return dispatch.concatenate(parts, axis=-1)


def pack_rfft(z: ArrayOrTensor, full_fft: bool = False) -> ArrayOrTensor:
    """
    Transform a real vector with :code:`size` elements to a vector of complex Fourier coefficients
    with :code:`size // 2 + 1` elements ready for inverse real fast Fourier transformation.

    Args:
        z: Unpacked vector of :code:`size` elements. See :func:`unpack_rfft` for details.
        full_fft: Whether to return the full set of Fourier coefficients rather than just the
            reduced representation for the real fast Fourier transform. The full representation is
            required for :func:`pack_rfft2`.

    Returns:
        rfft: Real Fourier transform coefficients.
    """
    *_, size = z.shape
    fftsize = size // 2 + 1
    ncomplex = (size - 1) // 2
    # Zero frequency term, real parts of complex coefficients and possible Nyqvist frequency.
    rfft = z[..., :fftsize] * (1 + 0j)
    # Imaginary parts of complex coefficients.
    rfft[..., 1:ncomplex + 1] += 1j * z[..., fftsize:]
    if not full_fft:
        return rfft
    # Add the redundant complex coefficients (use `flip` because torch does not support negative
    # strides).
    return dispatch.concatenate([rfft, dispatch.flip(rfft[..., 1:ncomplex + 1].conj(), (-1,))], -1)


def transform_irfft(z: ArrayOrTensor, loc: ArrayOrTensor, *, cov_rfft: OptionalArrayOrTensor = None,
                    cov: OptionalArrayOrTensor = None, rfft_scale: OptionalArrayOrTensor = None) \
        -> ArrayOrTensor:
    """
    Transform white noise in the Fourier domain to a Gaussian process realization.

    Args:
        z: Fourier-domain white noise with shape :code:`(..., size)`. See :func:`unpack_rfft` for
            details.
        loc: Mean of the Gaussian process with shape :code:`(..., size)`.
        cov_rfft: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., size // 2 + 1)`.
        cov: First row of the covariance matrix with shape :code:`(..., size)`.
        rfft_scale: Precomputed real fast Fourier transform scale with shape
            :code:`(..., size // 2 + 1)`.

    Returns:
        y: Realization of the Gaussian process with shape :code:`(..., size)`.
    """
    rfft_scale = _get_rfft_scale(cov_rfft, cov, rfft_scale, z.shape[-1])
    rfft = pack_rfft(z) * rfft_scale
    return dispatch[rfft].fft.irfft(rfft, z.shape[-1]) + loc


def transform_rfft(y: ArrayOrTensor, loc: ArrayOrTensor, *, cov_rfft: OptionalArrayOrTensor = None,
                   cov: OptionalArrayOrTensor = None, rfft_scale: OptionalArrayOrTensor = None) \
        -> ArrayOrTensor:
    """
    Transform a Gaussian process realization to white noise in the Fourier domain.

    Args:
        y: Realization of the Gaussian process with shape :code:`(..., size)`.
        loc: Mean of the Gaussian process with shape :code:`(..., size)`.
        cov_rfft: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., size // 2 + 1)`.
        cov: First row of the covariance matrix with shape :code:`(..., size)`.
        rfft_scale: Precomputed real fast Fourier transform scale with shape
            :code:`(..., size // 2 + 1)`.

    Returns:
        z: Fourier-domain white noise with shape :code:`(..., size)`. See :func:`transform_irrft`
            for details.
    """
    rfft_scale = _get_rfft_scale(cov_rfft, cov, rfft_scale, size=y.shape[-1])
    return unpack_rfft(dispatch[y].fft.rfft(y - loc) / rfft_scale, y.shape[-1])


def evaluate_log_prob_rfft(y: ArrayOrTensor, loc: ArrayOrTensor, *,
                           cov_rfft: OptionalArrayOrTensor = None,
                           cov: OptionalArrayOrTensor = None,
                           rfft_scale: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Evaluate the log probability of a one-dimensional Gaussian process realization in Fourier space.

    Args:
        y: Realization of a Gaussian process with shape :code:`(..., size)`, where :code:`...` is
            the batch shape and :code:`size` is the number of grid points.
        loc: Mean of the Gaussian process with shape :code:`(..., size)`.
        cov_rfft: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., size // 2 + 1)`.
        cov: First row of the covariance matrix with shape :code:`(..., size)`.
        rfft_scale: Precomputed real fast Fourier transform scale with shape
            :code:`(..., size // 2 + 1)`.

    Returns:
        log_prob: Log probability of the Gaussian process realization with shape :code:`(...)`.
    """
    rfft_scale = _get_rfft_scale(cov_rfft, cov, rfft_scale, y.shape[-1])
    rfft = transform_rfft(y, loc, rfft_scale=rfft_scale)
    return log_prob_stdnorm(rfft).sum(axis=-1) \
        + evaluate_rfft_log_abs_det_jacobian(y.shape[-1], rfft_scale=rfft_scale)


def evaluate_rfft_log_abs_det_jacobian(size: int, *, cov_rfft: OptionalArrayOrTensor = None,
                                       cov: OptionalArrayOrTensor = None,
                                       rfft_scale: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Evaluate the log absolute determinant of the Jacobian associated with :func:`transform_rfft`.

    Args:
        cov_rfft: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., size // 2 + 1)`.
        cov: First row of the covariance matrix with shape :code:`(..., size)`.
        rfft_scale: Precomputed real fast Fourier transform scale with shape
            :code:`(..., size // 2 + 1)`.

    Returns:
        log_abs_det_jacobian: Log absolute determinant of the Jacobian.
    """
    imagidx = (size + 1) // 2
    rfft_scale = _get_rfft_scale(cov_rfft, cov, rfft_scale, size)
    assert rfft_scale.shape[-1] == size // 2 + 1
    return - dispatch.log(rfft_scale).sum(axis=-1) \
        - dispatch.log(rfft_scale[1:imagidx]).sum(axis=-1) - log2 * ((size - 1) // 2) \
        + size * math.log(size) / 2



from .. import ArrayOrTensor


sqrt2 = 1.4142135623730951
log2 = 0.6931471805599453
log2pi = 1.8378770664093453


def log_prob_stdnorm(y: ArrayOrTensor) -> ArrayOrTensor:
    """
    Evaluate the log probability of a standard normal random variable.
    """
    return - (log2pi + y * y) / 2



from gptools.util import graph
import networkx as nx
import numbers
import numpy as np
import pytest
import re
from typing import Literal, Optional, Tuple, Union


@pytest.mark.parametrize("shape, ks", [
    ((23,), 5),
    ((19, 23), 7),
    ((19, 23), (5, 7)),
    ((7, 9, 11), 3),
    ((7, 9, 11), (3, 4, 5)),
])
@pytest.mark.parametrize("bounds", graph.LatticeBounds)
@pytest.mark.parametrize("compress", [False, True])
def test_lattice_predecessors(
        shape: Tuple[int], ks: Union[int, Tuple[int]], bounds: graph.LatticeBounds,
        compress: bool) -> None:
    predecessors = graph.lattice_predecessors(shape, ks, bounds, compress)

    # General shape check.
    rows, cols = predecessors.shape
    assert rows == np.prod(shape)
    expected_cols = np.prod(2 * np.asarray(ks) * np.ones_like(shape) + 1) - 1
    if compress:
        assert cols < expected_cols
    else:
        assert cols == expected_cols

    # Shape check for up to two dimensions if `k` is a scalar and the predecessors are compressed.
    if isinstance(ks, numbers.Number) and len(shape) < 3 and compress:
        expected_cols = graph.num_lattice_predecessors(ks, bounds, len(shape))
        assert cols == expected_cols


@pytest.mark.parametrize("shape, ks, bounds, match", [
    ((5, 6), (3, 7), "cube", "exceeds the tensor size"),
    ((10, 6), 2, "invalid-shape", "'invalid-shape' is not a valid LatticeBounds"),
])
def test_lattice_predecessors_invalid(
        shape: Tuple[int], ks: Tuple[int], bounds: graph.LatticeBounds, match: str) -> None:
    with pytest.raises(ValueError, match=re.escape(match)):
        graph.lattice_predecessors(shape, ks, bounds=bounds)


@pytest.mark.parametrize("indexing", ["numpy", "stan"])
def test_predecessors_to_edge_index(indexing: Literal["numpy", "stan"]) -> None:
    shape = (5, 7, 11)
    ks = (2, 3, 5)
    predecessors = graph.lattice_predecessors(shape, ks)
    edge_index = graph.predecessors_to_edge_index(predecessors, indexing)
    # The number of edges is the number of all combinations less the neighbors that fall outside the
    # bounds of the tensor. We just do a weak check here.
    assert edge_index.shape[0] == 2
    assert edge_index.shape[1] < predecessors.size
    # Check the validity of the edge index obtained from spatial predecessors.
    assert graph.check_edge_index(edge_index, indexing=indexing) is edge_index


@pytest.mark.parametrize("predecessors, match", [
    (np.asarray([[0], [0]]), "self-loops are not allowed; found 1"),
    (np.arange(3), "must be a matrix"),
])
def test_predecessors_to_edge_index_invalid(predecessors: np.ndarray, match: str) -> None:
    with pytest.raises(ValueError, match=match):
        graph.predecessors_to_edge_index(predecessors)


@pytest.mark.parametrize("edge_index, indexing, match", [
    (np.zeros(2), None, "edge index must have shape (2, num_edges) but got (2,)"),
    (np.zeros((3, 2)), None, "edge index must have shape (2, num_edges) but got (3, 2)"),
    (np.zeros((2, 3)), "foobar", "`indexing` must be one of"),
    (np.zeros((2, 3)), "stan", "expected indexing to start at 1"),
    (-np.ones((2, 3)), "numpy", "expected indexing to start at 0"),
    ([np.ones(2), np.ones(2)], "numpy", "self-loops are not allowed"),
    ([[0, 1, 2], [1, 2, 0]], "numpy", "successor indices must be non-decreasing"),
    ([[1], [0]], "numpy", "predecessors must be less than successors"),
])
def test_check_edge_index_invalid(
        edge_index: np.ndarray, indexing: Literal["numpy", "stan"],
        match: Optional[str]) -> None:
    with pytest.raises(ValueError, match=re.escape(match)):
        graph.check_edge_index(np.asarray(edge_index), indexing)


def test_compress_predecessors_invalid_shape():
    with pytest.raises(ValueError):
        graph.compress_predecessors(np.zeros(2))


@pytest.mark.parametrize("return_mapping", [False, True])
def test_edge_index_to_graph_roundtrip(return_mapping: bool) -> None:
    G = nx.erdos_renyi_graph(10, 0.1)
    edge_index = graph.graph_to_edge_index(G, indexing="numpy", return_mapping=return_mapping)
    if return_mapping:
        edge_index, mapping = edge_index
        assert isinstance(mapping, dict)
    H = graph.edge_index_to_graph(edge_index)
    assert set(map(tuple, G.edges)) == set(map(tuple, H.edges))



from gptools.util import ArrayOrTensorDispatch, coordgrid, fft, kernels
import numpy as np
import pytest
from scipy import stats
import torch as th
from typing import Tuple

rfft2_shapes = [(4, 6), (4, 7), (5, 6), (5, 7)]


@pytest.fixture(params=rfft2_shapes, ids=["-".join(map(str, shape)) for shape in rfft2_shapes])
def rfft2_shape(request: pytest.FixtureRequest) -> Tuple[int]:
    return request.param


@pytest.fixture(params=[4, 5, 9, 10])
def rfft_num(request: pytest.FixtureRequest) -> int:
    return request.param


@pytest.fixture(params=[(), (11,), (13, 11)])
def batch_shape(request: pytest.FixtureRequest) -> Tuple[int]:
    return request.param


def test_log_prob_norm(use_torch: bool) -> None:
    x = th.randn([]) if use_torch else np.random.normal()
    np.testing.assert_allclose(stats.norm(0, 1).logpdf(x), fft.log_prob_stdnorm(x))


def test_evaluate_log_prob_rfft(batch_shape: Tuple[int], rfft_num: int, use_torch: bool) -> None:
    x = np.linspace(0, 1, rfft_num, endpoint=False)
    kernel = kernels.ExpQuadKernel(np.random.gamma(10, 0.1), np.random.gamma(10, 0.01), 1) \
        + kernels.DiagonalKernel(0.1, 1)
    cov = kernel.evaluate(x[:, None])
    loc = np.random.normal(0, 1, rfft_num)
    dist = stats.multivariate_normal(loc, cov)
    y = dist.rvs(batch_shape)
    log_prob = dist.logpdf(y)
    lincov = cov[0]
    if use_torch:
        y = th.as_tensor(y)
        loc = th.as_tensor(loc)
        lincov = th.as_tensor(lincov)
    log_prob_rfft = fft.evaluate_log_prob_rfft(y, loc, cov=lincov)
    np.testing.assert_allclose(log_prob, log_prob_rfft)
    cov_rfft = ArrayOrTensorDispatch()[cov].fft.rfft(lincov)
    log_prob_rfft = fft.evaluate_log_prob_rfft(y, loc, cov_rfft=cov_rfft)
    np.testing.assert_allclose(log_prob, log_prob_rfft)


def test_transform_rfft_roundtrip(batch_shape: Tuple[int], rfft_num: int, use_torch: bool) -> None:
    x = np.linspace(0, 1, rfft_num, endpoint=False)
    kernel = kernels.ExpQuadKernel(np.random.gamma(10, 0.1), np.random.gamma(10, 0.01), 1) \
        + kernels.DiagonalKernel(0.1, 1)
    z = np.random.normal(0, 1, (*batch_shape, rfft_num))
    loc = np.random.normal(0, 1, rfft_num)
    cov = kernel.evaluate(x[:, None])
    cov = cov[0]
    if use_torch:
        loc = th.as_tensor(loc)
        z = th.as_tensor(z)
        cov = th.as_tensor(cov)
    y = fft.transform_irfft(z, loc, cov=cov)
    x = fft.transform_rfft(y, loc, cov=cov)
    # Verify that the inverse of the transform is the input.
    np.testing.assert_allclose(z, x)


@pytest.mark.parametrize("kernel", [
    kernels.ExpQuadKernel(np.random.gamma(10, 0.1), np.random.gamma(10, 0.01), 1.2),
    kernels.MaternKernel(1.5, np.random.gamma(10, 0.1), np.random.gamma(10, 0.01), 1.3),
    kernels.MaternKernel(2.5, np.random.gamma(10, 0.1), np.random.gamma(10, 0.01), 1.4),
])
def test_evaluate_log_prob_rfft2(kernel: kernels.Kernel, batch_shape: Tuple[int], rfft2_shape: int,
                                 use_torch: bool) -> None:
    xs = coordgrid(*(np.linspace(0, kernel.period, size, endpoint=False) for size in rfft2_shape))
    kernel = kernel + kernels.DiagonalKernel(1e-2, kernel.period)
    cov = kernel.evaluate(xs)
    loc = np.random.normal(0, 1, xs.shape[0])
    dist = stats.multivariate_normal(loc, cov)
    y = dist.rvs(batch_shape)
    log_prob = dist.logpdf(y)
    y2 = y.reshape(batch_shape + rfft2_shape)
    loc2 = loc.reshape(rfft2_shape)
    cov2 = cov[0].reshape(rfft2_shape)
    if use_torch:
        y2 = th.as_tensor(y2)
        loc2 = th.as_tensor(loc2)
        cov2 = th.as_tensor(cov2)
    log_prob_rfft2 = fft.evaluate_log_prob_rfft2(y2, loc2, cov=cov2)
    np.testing.assert_allclose(log_prob, log_prob_rfft2)
    cov_rfft2 = ArrayOrTensorDispatch()[cov].fft.rfft2(cov2)
    log_prob_rfft = fft.evaluate_log_prob_rfft2(y2, loc2, cov_rfft2=cov_rfft2)
    np.testing.assert_allclose(log_prob, log_prob_rfft)


def test_pack_rfft2_roundtrip(batch_shape: Tuple[int], rfft2_shape: int, use_torch: bool) -> None:
    z = np.random.normal(0, 1, batch_shape + rfft2_shape)
    if use_torch:
        z = th.as_tensor(z)

    # Unpacked Fourier to Fourier and back.
    y = fft.fft2.pack_rfft2(z)
    x = fft.fft2.unpack_rfft2(y, z.shape)
    np.testing.assert_allclose(z, x)

    # Fourier to unpacked Fourier and back.
    dispatch = ArrayOrTensorDispatch()
    z = dispatch[z].fft.rfft2(z)
    y = fft.fft2.unpack_rfft2(z, rfft2_shape)
    x = fft.fft2.pack_rfft2(y)
    np.testing.assert_allclose(z, x)


def test_transform_rfft2_roundtrip(batch_shape: Tuple[int], rfft2_shape: int, use_torch: bool) \
        -> None:
    xs = coordgrid(*(np.linspace(0, 1, size, endpoint=False) for size in rfft2_shape))
    kernel = kernels.ExpQuadKernel(np.random.gamma(10, 0.1), np.random.gamma(10, 0.01), 1) \
        + kernels.DiagonalKernel(1e-3, 1)
    cov = kernel.evaluate(xs)
    cov = cov[0].reshape(rfft2_shape)
    loc = np.random.normal(0, 1, rfft2_shape)
    z = np.random.normal(0, 1, batch_shape + rfft2_shape)
    if use_torch:
        z = th.as_tensor(z)
        cov = th.as_tensor(cov)
    y = fft.transform_irfft2(z, loc, cov=cov)
    x = fft.transform_rfft2(y, loc, cov=cov)
    # Verify that the inverse of the transform is the input.
    np.testing.assert_allclose(z, x)


@pytest.mark.parametrize("n", [5, 7])
def test_expand_rfft(n: int) -> None:
    x = np.random.normal(0, 1, n)
    rfft = np.fft.rfft(x)
    np.testing.assert_allclose(np.fft.fft(x), fft.expand_rfft(rfft, n))


@pytest.mark.parametrize("n", [21, 22])
def test_rfft_log_prob_pseudocode(n: int) -> None:
    f, loc = np.random.normal(0, 1, (2, n))
    cov = kernels.ExpQuadKernel(1.2, 0.1, n).evaluate(np.arange(n)[:, None])
    cov_rfft = np.fft.rfft(cov[0]).real
    desired = fft.evaluate_log_prob_rfft(f, loc, cov_rfft=cov_rfft)

    z = np.abs(np.fft.rfft(f - loc) / np.sqrt(n))
    actual = stats.norm(0, cov_rfft[0] ** 0.5).logpdf(z[0])
    if n % 2:
        m = (n + 1) // 2
    else:
        m = n // 2
        actual += stats.norm(0, cov_rfft[m] ** 0.5).logpdf(z[m])
    actual += 2 * stats.norm(0, cov_rfft[1:m] ** 0.5).logpdf(z[1:m]).sum()

    np.testing.assert_allclose(actual, desired)

    # Let's also verify that this matches the actual log prob.
    np.testing.assert_allclose(actual, stats.multivariate_normal(loc, cov).logpdf(f).sum())


@pytest.mark.parametrize("n", [5, 8])
def test_rfft_inv_pseudocode(n: int) -> None:
    z, loc = np.random.normal(0, 1, (2, n))
    cov_rfft = kernels.ExpQuadKernel(1.2, 0.3, n).evaluate_rfft(n)
    desired = fft.transform_irfft(z, loc, cov_rfft=cov_rfft)

    ftilde = np.zeros(n // 2 + 1, dtype=complex)
    # Zero-frequency and Nyquist terms.
    scale = (n * cov_rfft) ** 0.5
    ftilde[0] = z[0] * scale[0]
    if n % 2:
        m = (n + 1) // 2
    else:
        m = n // 2
        ftilde[m] = z[m] * scale[m]
    # Complex terms.
    ftilde[1:m] = scale[1:m] * (z[1:m] + 1j * z[m + (n + 1) % 2:n]) / np.sqrt(2)

    actual = np.fft.irfft(ftilde, n) + loc

    np.testing.assert_allclose(actual, desired)


@pytest.mark.parametrize("n", [9, 10])
def test_orthonormal(n: int) -> None:
    x = np.random.normal(0, 1, n)
    y = np.fft.fft(x)

    # Manually verify the transform using matrix multiplication.
    M = np.exp(-2 * np.pi * 1j * np.arange(n) * np.arange(n)[:, None] / n)
    Minv = np.exp(2 * np.pi * 1j * np.arange(n) * np.arange(n)[:, None] / n) / n
    z = M @ x
    np.testing.assert_allclose(y, z)
    np.testing.assert_allclose(x.real, (Minv @ y).real)



from gptools.util import plotting
import numpy as np


def test_plot_band() -> None:
    x = np.linspace(0, 1, 21)
    ys = np.random.normal(0, 1, (43, 21))
    line, band = plotting.plot_band(x, ys)




import enum
import numpy as np
from typing import Literal, Tuple, Union
from . import coordgrid, FALSE


if FALSE:
    import networkx as nx


# Solutions to the Gauss circle problem (https://en.wikipedia.org/wiki/Gauss_circle_problem).
GAUSS_PROBLEM_SEQUENCE = np.asarray([1, 5, 13, 29, 49, 81, 113, 149, 197, 253, 317, 377, 441])


class LatticeBounds(enum.Enum):
    """
    Boundary shape for the receptive field on a lattice.

    - CUBE results in a hypercuboid with dimensions :code:`2 * k + 1`.
    - DIAMOND results in the highest-volume hypercuboid whose vertices are axis-aligned that can fit
        into CUBE.
    - ELLIPSE results in the highest-volume ellipsoid that can fit into CUBE.
    """
    CUBE = "cube"
    DIAMOND = "diamond"
    ELLIPSE = "ellipse"


def lattice_predecessors(
        shape: Tuple[int], k: Union[int, Tuple[int]],
        bounds: LatticeBounds = LatticeBounds.ELLIPSE, compress: bool = True
        ) -> np.ndarray:
    """
    Evaluate predecessors for nodes on a lattice with given window size. Approximations will
    generally be poor if the half window width is smaller than the correlation length of the kernel.

    Args:
        shape: Shape of the tensor with Gaussian process distribution.
        k: Half window width or sequence of half window widths for each dimension. Each node will
            have a receptive field at most :code:`k` to the "left" and "right" in each dimension.
        bounds: Bounds of the receptive field. See :class:`LatticeBounds` for details.
        compress: Whether to compress predecessors such that the number of colums is equal to the
            maximum degree.

    Returns:
        predecessors: Mapping from each element of the tensor to its predecessors with shape
            :code:`(prod(shape), l)`, where :code:`l` is the number of predecessors.
    """
    bounds = LatticeBounds(bounds)
    # Convert shape and window widths to arrays and verify bounds.
    shape = np.asarray(shape)
    k = k * np.ones_like(shape)
    for i, (size, width) in enumerate(zip(shape, 2 * k + 1)):
        if width > size:
            raise ValueError(f"window width {width} exceeds the tensor size {size} along dim {i}")
    # Get all possible combinations of indices and steps.
    coords = coordgrid(*[np.arange(p) for p in shape], ravel=True)
    steps = coordgrid(*[np.roll(np.arange(- s, s + 1), - s) for s in k], ravel=True)
    # Construct the vector predecessors and a mask that removes indices outside the tensor bounds.
    predecessors = coords[..., None, :] - steps
    assert predecessors.shape == (shape.prod(), (2 * k + 1).prod(), len(shape))
    mask = ((predecessors >= 0) & (predecessors < shape)).all(axis=-1)
    # If we want an ellipsoidal predecessors, we remove ones that are too far away.
    if bounds == LatticeBounds.ELLIPSE:
        t = np.square(steps / k).sum(axis=-1)
        mask &= t <= 1
    elif bounds == LatticeBounds.DIAMOND:
        t = np.abs(steps / k).sum(axis=-1)
        mask &= t <= 1
    # Mask invalid indices, ravel the coordinate indices, and mask again after ravelling.
    predecessors = np.where(mask[..., None], predecessors, 0)
    predecessors = np.ravel_multi_index(np.moveaxis(predecessors, -1, 0), shape)
    # Ensure all nodes are predecessors so the graph is acyclic.
    mask &= predecessors <= np.arange(coords.shape[0])[:, None]
    predecessors = np.where(mask, predecessors, -1)[:, 1:]
    if compress:
        predecessors = compress_predecessors(predecessors)
    return predecessors


def num_lattice_predecessors(k: int, bounds: LatticeBounds, p: int) -> int:
    """
    Evaluate the number of predecessors in a lattice graph.
    """
    bounds = LatticeBounds(bounds)
    if p == 1:
        return k
    elif p == 2 and bounds == LatticeBounds.CUBE:
        return 2 * k * (k + 1)
    elif p == 2 and bounds == LatticeBounds.DIAMOND:
        return k * (k + 1)
    elif p == 2 and bounds == LatticeBounds.ELLIPSE:
        return (GAUSS_PROBLEM_SEQUENCE[k] - 1) // 2
    else:
        raise NotImplementedError(f"k = {k}; bounds = {bounds}; p = {p}")


def compress_predecessors(predecessors: np.ndarray) -> np.ndarray:
    """
    Compress a predecessor matrix such that there is at least one predecessors that does not contain
    invalid indices. In other words, we remove as many columns as possible without discarding any
    information.

    Args:
        predecessors: Mapping from each element of the tensor to its predecessors.

    Returns:
        compressed: Predecessors after removing as many columns as possible.
    """
    if predecessors.ndim != 2:
        raise ValueError("predecessors must be a matrix")
    num_nodes, _ = predecessors.shape
    max_degree = (predecessors >= 0).sum(axis=1).max()
    compressed = - np.ones((num_nodes, max_degree), dtype=predecessors.dtype)
    for i, neighbors in enumerate(predecessors):
        neighbors = neighbors[neighbors >= 0]
        compressed[i, :len(neighbors)] = neighbors
    return compressed


def _check_indexing(indexing: Literal["numpy", "stan"]) -> Literal["numpy", "stan"]:
    if indexing not in (expected := {"numpy", "stan"}):
        raise ValueError(f"`indexing` must be one of {expected} but got {indexing}")
    return indexing


def predecessors_to_edge_index(predecessors: np.ndarray,
                               indexing: Literal["numpy", "stan"] = "stan") -> np.ndarray:
    """
    Convert a matrix of predecessors to an edgelist with self loops.

    Args:
        predecessors: Predecessor matrix such that each row corresponds to the predecessors of the
            associated node. Negative node labels are omitted.
        indexing: Whether to use zero-based indexing (`numpy`) or one-based indexing (`stan`).

    Returns:
        edge_index: Tuple of parent and child node labels.
    """
    if predecessors.ndim != 2:
        raise ValueError("predecessors must be a matrix")
    nodes = np.arange(predecessors.shape[0])
    if num_self_loops := (predecessors == nodes[:, None]).sum():
        raise ValueError(f"self-loops are not allowed; found {num_self_loops}")

    edge_index = np.transpose([(parent, child) for child, parents in enumerate(predecessors) for
                               parent in parents if parent >= 0])
    if _check_indexing(indexing) == "stan":
        return edge_index + 1
    return edge_index


def check_edge_index(edge_index: np.ndarray, indexing: Literal["numpy", "stan"] = "stan") \
        -> np.ndarray:
    """
    Check the edge index, ensuring it has the right structure for the Stan implementation and
    corresponds to a directed acyclic graph.

    Args:
        edge_index: Tuple of parent and child node labels to check.
        indexing: Whether to use zero-based indexing (`numpy`) or one-based indexing (`stan`).

    Returns:
        edge_index: Edge index after checking.
    """
    # Check the tensor shape.
    if edge_index.ndim != 2 or edge_index.shape[0] != 2:
        raise ValueError(f"edge index must have shape (2, num_edges) but got {edge_index.shape}")

    # Check that child nodes start with the proper value.
    expected_min = 0 if _check_indexing(indexing) == "numpy" else 1
    if (actual_min := edge_index.min()) < expected_min:
        raise ValueError(f"expected indexing to start at {expected_min}; found {actual_min}")
    predecessors, successors = edge_index

    # Check that child labels are non-decreasing.
    if (np.diff(successors) < 0).any():
        raise ValueError("successor indices must be non-decreasing")

    # Check that there are no self-loops.
    if num_self_loops := (successors == predecessors).sum():
        raise ValueError(f"self-loops are not allowed; found {num_self_loops}")

    # Check that predecessors are less than successors.
    if (predecessors >= successors).any():
        raise ValueError("predecessors must be less than successors")

    return edge_index


def edge_index_to_graph(edge_index: np.ndarray) -> "nx.DiGraph":
    """
    Convert edge indices to a directed graph.

    Args:
        edge_index: Tuple of parent and child node labels.

    Returns:
        graph: Directed graph induced by the edge indices.
    """
    import networkx as nx
    graph = nx.DiGraph()
    graph.add_edges_from(edge_index.T)
    return graph


def graph_to_edge_index(graph: "nx.Graph", return_mapping: bool = False,
                        indexing: Literal["numpy", "stan"] = "stan") \
        -> Union[np.ndarray, Tuple[np.ndarray, dict]]:
    """
    Convert a graph to edge indices.

    Args:
        graph: Graph to convert.
        add_self_loops: Whether to include self loops.
        return_mapping: Whether to return the mapping from node labels to indices.
        indexing: Whether to use zero-based indexing (`numpy`) or one-based indexing (`stan`).

    Returns:
        edge_index: Tuple of parent and child node labels.
    """
    start = 1 if indexing == "stan" else 0
    mapping = {node: i for i, node in enumerate(sorted(graph), start=start)}
    edge_index = []
    for node in sorted(graph):
        neighbors = graph.neighbors(node)
        node = mapping[node]
        for neighbor in neighbors:
            neighbor = mapping[neighbor]
            if neighbor < node:
                edge_index.append((neighbor, node))

    edge_index = np.transpose(edge_index)
    if return_mapping:
        return edge_index, mapping
    return edge_index



import math
import numpy as np
import operator
from typing import Callable, Iterable, Tuple, Union
from . import ArrayOrTensor, ArrayOrTensorDispatch, coordgrid, OptionalArrayOrTensor
from .fft import expand_rfft


dispatch = ArrayOrTensorDispatch()


def evaluate_residuals(x: ArrayOrTensor, y: OptionalArrayOrTensor = None,
                       period: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Evaluate the residuals between points respecting periodic boundary conditions.

    If :code:`period is not None` and boundary conditions apply, residuals have the correct "local"
    behavior, i.e., points to the left have a negative residual and points to the right have a
    positive residual. This leads to a discontinuity a distance :code:`period / 2` from any
    reference point. The discontinuity is immaterial for even kernel functions.

    Args:
        x: Coordinates with shape :code:`(..., p)`, where :code:`...` is the batch shape and
            :code:`p` is the number of dimensions of the embedding space.
        y: Coordinates with shape :code:`(..., p)` which must be broadcastable to :code:`x`. If not
            given, the distance between the Cartesian product of :code:`x` will be evaluated.
        period: Period of circular boundary conditions.

    Returns:
        dist2: Squared distance between :code:`x` and :code:`y`.

    Example:
      .. plot::

        from gptools.util.kernels import evaluate_residuals
        from matplotlib import pyplot as plt
        import numpy as np

        width = 3  # Width of the domain.
        step_seq = [6, 7]  # Number of grid points in the domain.
        boundary_color = "silver"  # Color for domain boundaries.

        fig, axes = plt.subplots(2, 1, sharex=True, sharey=True)

        for ax, steps in zip(axes, step_seq):
            # Plot distances for two locations.
            x = np.linspace(-width, 2 * width, 3 * steps + 1, endpoint=True)
            ys = [x[steps + 2], x[2 * steps - 1]]
            for y in ys:
                line, = ax.plot(x, evaluate_residuals(x, y, width), marker=".")
                ax.scatter(y, 0, color=line.get_color(), label=f"$y={y:.1f}$")

            # Plot boundary indicators.
            for i in range(-1, 3):
                ax.axvline(i * width, color=boundary_color, ls="--")
            ax.axhline(width / 2, color=boundary_color, ls="--")
            ax.axhline(-width / 2, color=boundary_color, ls="--", label="domain boundaries")
            ax.axhline(0, color="silver", ls=":")

            ax.set_aspect("equal")
            ax.set_ylabel(r"residual $x - y$")
            ax.set_title(fr"$n={steps}$")

        axes[1].set_xlabel("position $x$")

        # Adjust boundaries.
        factor = 0.2
        ax.set_ylim(-(1 + factor) * width / 2, (1 + factor) * width / 2)
        ax.set_xlim(-(1 + factor / 2) * width, (2 + factor / 2) * width)
        fig.tight_layout()
    """
    # Expand the shape so we get the Cartesian product of elements in x (while keeping the batch
    # shape).
    if y is None:
        x, y = x[..., :, None, :], x[..., None, :, :]
    residuals = x - y
    if period is not None:
        residuals = residuals - period * dispatch.round(residuals / period)
    return residuals


def evaluate_squared_distance(x: ArrayOrTensor, y: OptionalArrayOrTensor = None,
                              period: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    r"""
    Evaluate the squared distance between points respecting periodic boundary conditions.

    Args:
        x: Coordinates with shape :code:`(..., p)`, where :code:`...` is the batch shape and
            :code:`p` is the number of dimensions of the embedding space.
        y: Coordinates with shape :code:`(..., p)` which must be broadcastable to :code:`x`. If not
            given, the distance between the Cartesian product of :code:`x` will be evaluated.
        period: Period of circular boundary conditions.

    Returns:
        dist2: Squared distance between :code:`x` and :code:`y`.

    Example:
      .. plot::

        from gptools.util import coordgrid
        from gptools.util.kernels import evaluate_squared_distance
        from matplotlib import pyplot as plt
        import numpy as np

        fig, axes = plt.subplots(1, 2, sharex=True, sharey=True)
        height = 40
        width = 50
        shape = (height, width)
        xs = coordgrid(np.arange(height), np.arange(width))

        idx = len(xs) // 4 + width // 4
        periods = [None, shape]
        dists = [np.sqrt(evaluate_squared_distance(xs[idx], xs, period=period))
                for period in periods]

        vmax = np.sqrt(width ** 2 + height ** 2)
        for ax, period, dist in zip(axes, periods, dists):
            dist = dist.reshape(shape)
            im = ax.imshow(dist, vmax=vmax, origin="lower")
            colorbar = fig.colorbar(im, ax=ax, location="top")
            label = r"distance $d\left(x,y\right)$"
            if period:
                label += " (periodic boundaries)"
            colorbar.set_label(label)
            cs = ax.contour(dist, colors="w", levels=[10, 20, 30], linestyles=["-", "--", ":"])
            plt.clabel(cs)
            ax.scatter(*np.unravel_index(idx, shape)[::-1], color="C1").set_edgecolor("w")
            ax.set_xlabel("position $x$")

        axes[0].set_ylabel("position $y$")
        fig.tight_layout()
    """
    residuals = evaluate_residuals(x, y, period)
    return (residuals * residuals).sum(axis=-1)


class Kernel:
    """
    Base class for covariance kernels.

    Args:
        period: Period for circular boundary conditions.
    """
    def __init__(self, period: OptionalArrayOrTensor = None):
        self.period = period

    def evaluate(self, x: ArrayOrTensor, y: OptionalArrayOrTensor = None) -> ArrayOrTensor:
        """
        Evaluate the covariance kernel.

        Args:
            x: First set of points.
            y: Second set of points (defaults to :code:`x` for pairwise covariances).

        Returns:
            cov: Covariance between the two sets of points.
        """
        raise NotImplementedError

    def evaluate_rfft(self, shape: Union[int, Tuple[int]]) -> ArrayOrTensor:
        """
        Evaluate the real fast Fourier transform of the kernel.

        Args:
            shape: Number of sample points in each dimension.

        Returns:
            rfft: Fourier coefficients with shape :code:`(*shape[:-1], shape[-1] // 2 + 1)`.
        """
        raise NotImplementedError

    def __add__(self, other) -> "CompositeKernel":
        return CompositeKernel(operator.add, self, other)

    @property
    def is_periodic(self):
        return self.period is not None


class CompositeKernel(Kernel):
    """
    Composition of two kernels.

    Args:
        operation: Operation for composing kernels.
        a: First kernel.
        b: Second kernel.
    """
    def __init__(self, operation: Callable, a: Kernel, b: Kernel) -> None:
        period = None
        if isinstance(a, Kernel) and isinstance(b, Kernel):
            if a.is_periodic != b.is_periodic:
                raise ValueError("either both or neither kernel must be periodic")
            if a.is_periodic:
                if not dispatch.allclose(a.period, b.period):
                    raise ValueError("kernels do not have the same period")
                period = a.period
        super().__init__(period)
        self.operation = operation
        self.a = a
        self.b = b

    def evaluate(self, x: ArrayOrTensor, y: OptionalArrayOrTensor = None) -> ArrayOrTensor:
        return self.operation(self.a.evaluate(x, y) if isinstance(self.a, Kernel) else self.a,
                              self.b.evaluate(x, y) if isinstance(self.b, Kernel) else self.b)


class DiagonalKernel(Kernel):
    """
    Diagonal kernel with "nugget" variance. The kernel can only evaluated pairwise for a single set
    of points but not for the Cartesian product of two sets of points.
    """
    def __init__(self, epsilon: float = 1, period: OptionalArrayOrTensor = None) -> None:
        super().__init__(period)
        self.epsilon = epsilon

    def evaluate(self, x: ArrayOrTensor, y: OptionalArrayOrTensor = None) -> ArrayOrTensor:
        if y is not None:
            raise ValueError
        return dispatch[x].eye(x.shape[-2]) * self.epsilon


class ExpQuadKernel(Kernel):
    r"""
    Exponentiated quadratic kernel or solution to the heat equation if the kernel is periodic.

    Args:
        sigma: Scale of the covariance.
        length_scale: Correlation length.
        period: Period for circular boundary conditions.
    """
    def __init__(self, sigma: float, length_scale: float, period: OptionalArrayOrTensor = None) \
            -> None:
        super().__init__(period)
        self.sigma = sigma
        self.length_scale = length_scale

    def evaluate(self, x: ArrayOrTensor, y: OptionalArrayOrTensor = None) -> ArrayOrTensor:
        residuals = evaluate_residuals(x, y, self.period) / self.length_scale
        exponent = - dispatch.square(residuals).sum(axis=-1) / 2
        return self.sigma * self.sigma * dispatch.exp(exponent)

    def evaluate_rfft(self, shape: Union[int, Tuple[int]]) -> ArrayOrTensor:
        if not self.is_periodic:
            raise NotImplementedError
        if not isinstance(shape, Iterable):
            shape = [shape]
        ndim = len(shape)
        rescaled_length_scale = self.length_scale * np.ones(ndim) / self.period
        value = None
        for i, size in enumerate(shape):
            xi = np.arange(size // 2 + 1)
            part = size * rescaled_length_scale[i] * (2 * math.pi) ** 0.5 \
                * np.exp(-2 * (math.pi * xi * rescaled_length_scale[i]) ** 2)
            if i != ndim - 1:
                part = expand_rfft(part, size)
            if value is None:
                value = part
            else:
                value = value[..., None] * part

        return self.sigma ** 2 * value


class MaternKernel(Kernel):
    """
    Matern covariance function.

    Args:
        dof: Smoothness parameter.
        sigma: Scale of the covariance.
        length_scale: Correlation length.
        period: Period for circular boundary conditions.
    """
    def __init__(self, dof: float, sigma: float, length_scale: float,
                 period: OptionalArrayOrTensor = None) -> None:
        super().__init__(period)
        if dof not in (allowed_dofs := {3 / 2, 5 / 2}):
            raise ValueError(f"dof must be one of {allowed_dofs} but got {dof}")
        self.sigma = sigma
        self.length_scale = length_scale
        self.dof = dof

    def evaluate(self, x: ArrayOrTensor, y: OptionalArrayOrTensor = None) -> ArrayOrTensor:
        residuals = evaluate_residuals(x, y, self.period) / self.length_scale
        distance = (2 * self.dof * residuals * residuals).sum(axis=-1) ** 0.5
        if self.dof == 3 / 2:
            value = 1 + distance
        elif self.dof == 5 / 2:
            value = 1 + distance + distance * distance / 3
        else:
            raise NotImplementedError
        return self.sigma * self.sigma * value * dispatch.exp(-distance)

    def evaluate_rfft(self, shape: Tuple[int]):
        if not self.is_periodic:
            raise NotImplementedError
        if not isinstance(shape, Iterable):
            shape = [shape]
        from scipy import special

        # Construct the grid to evaluate on.
        size = np.prod(shape)
        *head, tail = shape
        ks = [np.arange(n) for n in head]
        ks.append(np.arange(tail // 2 + 1))
        ks = coordgrid(*ks)
        ks = np.minimum(ks, shape - ks)
        ndim = len(shape)

        # Evaluate the spectral density.
        length_scale = self.length_scale / self.period * np.ones(ndim)
        arg = 1 + 2 / self.dof * np.sum((math.pi * length_scale * ks) ** 2, axis=-1)
        value = size * 2 ** ndim * (math.pi / (2 * self.dof)) ** (ndim / 2) \
            * special.gamma(self.dof + ndim / 2) / special.gamma(self.dof) \
            * arg ** -(self.dof + ndim / 2) * length_scale.prod()
        return self.sigma * self.sigma * value.reshape(head + [tail // 2 + 1])



import multiprocessing
import numbers
import psutil
from queue import Empty
import traceback
from typing import Any, Callable


def _wrapper(queue, target, *args, **kwargs):  # pragma: no cover
    """
    Wrapper to execute a function in a subprocess.
    """
    try:
        result = target(*args, **kwargs)
        success = True
    except Exception as ex:
        result = (ex, traceback.format_exc())
        success = False
    queue.put_nowait((success, result))


def call_with_timeout(timeout: float, target: Callable, *args, **kwargs) -> Any:
    """
    Call a target with a timeout and return its result.

    Args:
        timeout: Number of seconds to wait for a result.
        target: Function to call.
        *args: Positional arguments passed to :code:`target`.
        **kwargs: Keyword arguments passed to :code:`target`.

    Returns:
        result: Return value of :code:`target`.

    Raises:
        TimeoutError: If the target does not complete within the timeout.
    """
    if not isinstance(timeout, numbers.Number) or timeout <= 0:
        raise ValueError("timeout must be a positive number")
    if not callable(target):
        raise TypeError("target must be callable")
    queue = multiprocessing.Queue()
    process = multiprocessing.Process(target=_wrapper, args=(queue, target, *args), kwargs=kwargs,
                                      daemon=True)
    process.start()

    try:
        success, result = queue.get(timeout=timeout)
    except Empty:
        raise TimeoutError(f"failed to fetch result after {timeout} seconds")
    finally:
        if process.is_alive():
            # Kill the process and all its children (https://stackoverflow.com/a/4229404/1150961).
            children = psutil.Process(process.pid).children(recursive=True)
            for child in children:
                child.kill()  # pragma: no cover
            _, still_alive = psutil.wait_procs(children, timeout=3)
            if still_alive:
                raise RuntimeError("some processes are still alive")  # pragma: no cover
            process.kill()
        process.join(timeout=5)
    if not success:
        ex, tb = result
        raise RuntimeError(tb) from ex
    return result



from __future__ import annotations
import functools as ft
import inspect
import numpy as np
import os
import time
from typing import Any, Callable, Iterable, Literal, Optional, Tuple, Union


# This tricks pylance into thinking that the imports *could* happen for type checking.
FALSE = os.environ.get("f1571823-5638-473a-adb5-6f5efb0cb773")
if FALSE:
    from matplotlib.axes import Axes
    from matplotlib.colorbar import Colorbar
    import torch as th


ArrayOrTensor = Union[np.ndarray, "th.Tensor"]
OptionalArrayOrTensor = Optional[ArrayOrTensor]


class ArrayOrTensorDispatch:
    """
    Call the equivalent numpy or torch function based on the value of arguments.
    """
    def __getattr__(self, name: str) -> Callable:
        return ft.partial(self, name)

    def __call__(self, name, *args, **kwargs) -> Any:
        module = self[args + tuple(kwargs.values())]
        return getattr(module, name)(*args, **kwargs)

    def __getitem__(self, x: ArrayOrTensor) -> Any:
        if self.is_tensor(*x) if isinstance(x, tuple) else self.is_tensor(x):
            import torch
            return torch
        else:
            return np

    def is_tensor(self, *xs: Iterable[ArrayOrTensor]) -> bool:
        """
        Check if objects are tensors.

        Args:
            xs: Objects to check.

        Returns:
            is_tensor: :code:`True` if :code:`x` is a tensor; :code:`False` otherwise.

        Raises:
            ValueError: If the objects are a mixture of torch tensors and numpy arrays.
        """
        #
        try:
            import torch as th
        except ModuleNotFoundError:
            return False
        # None of the arguments are tensors.
        if not any(isinstance(x, th.Tensor) for x in xs):
            return False
        # At least one is a tensor. If another is an array, we're in trouble.
        if any(isinstance(x, np.ndarray) for x in xs):
            raise ValueError("arguments are a mixture of torch tensors and numpy arrays")
        return True

    def get_complex_dtype(self, x: ArrayOrTensor) -> Any:
        """
        Get the complex dtype matching :code:`x` in precision.
        """
        return (1j * self[x].empty(0, dtype=x.dtype)).dtype

    @ft.wraps(np.concatenate)
    def concatenate(self, arrays: Iterable[ArrayOrTensor],
                    axis: Optional[Union[int, Tuple[int]]] = None) -> ArrayOrTensor:
        if self.is_tensor(*arrays):
            import torch as th
            return th.concat(arrays, dim=axis)
        else:
            return np.concatenate(arrays, axis=axis)


dispatch = ArrayOrTensorDispatch()


def coordgrid(*xs: Iterable[np.ndarray], ravel: bool = True, indexing: Literal["ij", "xy"] = "ij") \
        -> np.ndarray:
    """
    Obtain coordinates for all grid points induced by :code:`xs`.

    Args:
        xs: Coordinates to construct the grid.
        ravel: Whether to reshape the leading dimensions.
        indexing: Whether to use Cartesian :code:`xy` or matrix :code:`ij` indexing (defaults to
            :code:`ij`).

    Returns:
        coord: Coordinates for all grid points with shape
            :code:`(len(xs[0]), ..., len(xs[p - 1]), p)` if :code:`ravel` is :code:`False`, where
            :code:`p = len(xs)` is the number of dimensions. If :code:`ravel` is :code:`True`, the
            shape is :code:`(len(xs[0]) * ... * len(xs[p - 1]), p)`.
    """
    # Stack the coordinate matrices and move the coordinate dimension to the back.
    coords = np.moveaxis(np.stack(np.meshgrid(*xs, indexing=indexing)), 0, -1)
    if not ravel:
        return coords
    return coords.reshape((-1, len(xs)))


class Timer:
    """
    Time the duration of code execution in a context.

    Args:
        message: Message to print when the context becomes inactive.
    """
    def __init__(self, message: str = None):
        self.start = None
        self.end = None
        self.message = message

    def __enter__(self) -> Timer:
        if self.start is not None:
            raise RuntimeError("timers can only be used once")
        self.start = time.time()
        return self

    def __exit__(self, *args) -> None:
        self.end = time.time()
        if self.message:
            print(f"{self.message} in {self.duration:.3f} seconds")

    @property
    def duration(self) -> float:
        """
        The duration between the start and end of the context. Returns the time since the start if
        the context is still active.
        """
        if self.start is None:
            raise RuntimeError("timer has not yet been started")
        end = self.end or time.time()
        return end - self.start

    def __repr__(self) -> str:
        try:
            return f"Timer(duration={self.duration:.3f})"
        except RuntimeError:
            return "Timer(not started)"


class mutually_exclusive_kwargs:
    """
    Ensure a function receives mutually exclusive keyword arguments.

    Args:
        keys: Sequence of string or string tuple keys. Tuples indicate that none or all of the keys
            must be given together.
    """
    def __init__(self, *keys) -> None:
        self.keys = keys

    def __call__(self, func: Callable) -> Callable:
        @ft.wraps(func)
        def _wrapper(*args, **kwargs) -> Any:
            # Assemble the values from the keys.
            given_key = None
            for key in self.keys:
                if isinstance(key, str):
                    given = kwargs.get(key) is not None
                else:
                    given = {x: kwargs.get(x) is not None for x in key}
                    if all(given.values()) != any(given.values()):
                        raise ValueError(f"some but not all of {key} are given: {given}")
                    given = all(given.values())
                if given and given_key:
                    raise ValueError(f"`{key}` and `{given_key}` are both given")
                if given:
                    given_key = key
            if not given_key:
                raise ValueError(f"expected exactly one of {self.keys} to be given but got "
                                 f"{kwargs}")
            if "given" in list(inspect.signature(func).parameters):
                kwargs["given"] = given_key
            return func(*args, **kwargs)
        return _wrapper


def encode_one_hot(z: ArrayOrTensor, p: Optional[int] = None) -> ArrayOrTensor:
    """
    Encode a vector of integers as a one-hot matrix.

    Args:
        z: Array of integers.
        p: Number of classes (defaults to :code:`max(z) + 1`).

    Returns:
        one_hot: One-hot encoded values.
    """
    p = p or z.max() + 1
    n = z.shape[0]
    result = dispatch[z].zeros((n, p))
    result[dispatch[z].arange(n), z] = 1
    return result


def match_colorbar(cb: "Colorbar", ax: Optional["Axes"] = None) -> Tuple[float]:
    """
    Match the size of the colorbar with the size of the axes.

    Args:
        ax: Axes from which the colorbar "stole" space.
        cb: Colorbar to match to :code:`ax`.

    Returns:
        pos: New position of the colorbar axes.
    """
    from matplotlib import pyplot as plt

    ax = ax or plt.gca()
    bbox = ax.get_position()
    cb_bbox = cb.ax.get_position()
    cb.ax.set_aspect("auto")
    if cb.orientation == "vertical":
        # Update bottom and height.
        left = cb_bbox.xmin
        width = cb_bbox.width
        bottom = bbox.ymin
        height = bbox.height
    else:
        # Update left and width.
        left = bbox.xmin
        width = bbox.width
        bottom = cb_bbox.ymin
        height = cb_bbox.height
    pos = (left, bottom, width, height)
    cb.ax.set_position(pos)
    return pos



import math
from typing import Optional, Tuple
from .. import ArrayOrTensor, ArrayOrTensorDispatch, mutually_exclusive_kwargs, \
    OptionalArrayOrTensor
from .fft1 import pack_rfft, unpack_rfft
from .util import log2, log_prob_stdnorm


dispatch = ArrayOrTensorDispatch()


def _get_rfft2_scale(cov_rfft2: OptionalArrayOrTensor, cov: OptionalArrayOrTensor,
                     rfft2_scale: OptionalArrayOrTensor, width: Optional[int]) -> ArrayOrTensor:
    num_given = sum([cov_rfft2 is not None, cov is not None, rfft2_scale is not None])
    if num_given != 1:  # pragma: no cover
        raise ValueError("exactly one of `cov_rfft2`, `cov`, or `rfft2_scale` must be given")
    if rfft2_scale is not None:
        return rfft2_scale
    return evaluate_rfft2_scale(cov_rfft2=cov_rfft2, cov=cov, width=width)


@mutually_exclusive_kwargs("cov", "cov_rfft2")
def evaluate_rfft2_scale(*, cov_rfft2: OptionalArrayOrTensor = None,
                         cov: OptionalArrayOrTensor = None, width: Optional[int] = None) \
        -> ArrayOrTensor:
    """
    Evaluate the scale of Fourier coefficients.

    Args:
        cov_rfft2: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., height, width // 2 + 1)`.
        cov: Covariance between the first grid point and the remainder of the grid with shape
            :code:`(..., height, width)`.
        width: Number of columns of the signal (cannot be inferred from the Fourier coefficients).

    Returns:
        scale: Scale of Fourier coefficients with shape :code:`(..., height, width // 2 + 1)`.
    """
    if cov is not None:
        *_, height, width = cov.shape
        cov_rfft2 = dispatch[cov].fft.rfft2(cov).real
    else:
        *_, height, _ = cov_rfft2.shape
    size = width * height
    rfft2_scale = size * cov_rfft2 / 2

    # Recall how the two-dimensional RFFT is computed. We first take an RFFT of rows of the matrix.
    # This leaves us with a real first column (zero frequency term) and a real last column if the
    # number of columns is even (Nyqvist frequency term). Second, we take a *full* FFT of the
    # columns. The first column will have a real coefficient in the first row (zero frequency in the
    # "row-dimension"). All elements in rows beyond n // 2 + 1 are irrelevant because the column was
    # real. The same applies to the last column if there is a Nyqvist frequency term. Finally, we
    # will also have a real-only Nyqvist frequency term in the first and last column if the number
    # of rows is even.

    # The zero-frequency term in both dimensions which must always be real.
    rfft2_scale[..., 0, 0] *= 2

    # If the number of colums is even, the last row will be real after the row-wise RFFT.
    # Consequently, the first element is real after the column-wise FFT.
    if width % 2 == 0:
        rfft2_scale[..., 0, width // 2] *= 2

    # If the number of rows is even, we have a Nyqvist frequency term in the first column.
    if height % 2 == 0:
        rfft2_scale[..., height // 2, 0] *= 2

    # If the number of rows and columns is even, we also have a Nyqvist frequency term in the last
    # column.
    if height % 2 == 0 and width % 2 == 0:
        rfft2_scale[..., height // 2, width // 2] *= 2

    return dispatch.sqrt(rfft2_scale)


def unpack_rfft2(z: ArrayOrTensor, shape: Tuple[int]) -> ArrayOrTensor:
    """
    Unpack the Fourier coefficients of a two-dimensional real Fourier transform with shape
    :code:`(..., height, width // 2 + 1)` to a batch of matrices with shape
    :code:`(..., height, width)`.

    TODO: add details on packing structure.

    Args:
        z: Two-dimensional real Fourier transform coefficients.
        shape: Shape of the real signal. Necessary because the number of columns cannot be inferred
            from :code:`rfft2`.

    Returns:
        z: Unpacked matrices with shape :code:`(..., height, width)`.
    """
    *_, height, width = shape
    ncomplex = (width - 1) // 2
    parts = [
        # First column is always real.
        unpack_rfft(z[..., :height // 2 + 1, 0], height)[..., None],
        # Real and imaginary parts of complex coefficients.
        z[..., 1:ncomplex + 1].real,
        z[..., 1:ncomplex + 1].imag,
    ]
    if width % 2 == 0:  # Nyqvist frequency terms if the number of columns is even.
        parts.append(unpack_rfft(z[..., :height // 2 + 1, width // 2], height)[..., None])
    return dispatch.concatenate(parts, axis=-1)


def pack_rfft2(z: ArrayOrTensor) -> ArrayOrTensor:
    """
    Transform a batch of real matrices with shape :code:`(..., height, width)` to a batch of complex
    Fourier coefficients with shape :code:`(..., height, width // 2 + 1)` ready for inverse real
    fast Fourier transformation in two dimensions.

    Args:
        z: Unpacked matrices with shape :code:`(..., height, width)`.

    Returns:
        rfft: Two-dimensional real Fourier transform coefficients.
    """
    *batch_shape, height, width = z.shape
    ncomplex = (width - 1) // 2
    rfft2 = dispatch[z].empty((*batch_shape, height, width // 2 + 1),
                              dtype=dispatch.get_complex_dtype(z))
    # Real FFT in the first column due to zero-frequency terms for the row-wise Fourier transform.
    rfft2[..., 0] = pack_rfft(z[..., 0], full_fft=True)
    # Complex Fourier coefficients.
    rfft2[..., 1:ncomplex + 1] = z[..., 1:ncomplex + 1] + 1j * z[..., ncomplex + 1:2 * ncomplex + 1]
    # Real FFT in the last column due to the Nyqvist frequency terms for the row-wise Fourier
    # transform if the number of columns is even.
    if width % 2 == 0:
        rfft2[..., width // 2] = pack_rfft(z[..., width - 1], full_fft=True)
    return rfft2


def transform_irfft2(z: ArrayOrTensor, loc: ArrayOrTensor, *,
                     cov_rfft2: OptionalArrayOrTensor = None, cov: OptionalArrayOrTensor = None,
                     rfft2_scale: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Transform white noise in the Fourier domain to a Gaussian process realization.

    Args:
        z: Unpacked matrices with shape :code:`(..., height, width)`. See :func:`unpack_rfft2` for
            details.
        loc: Mean of the Gaussian process with shape :code:`(..., height, width)`.
        cov_rfft2: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., height, width // 2 + 1)`.
        cov: Covariance between the first grid point and the remainder of the grid with shape
            :code:`(..., height, width)`.
        rfft2_scale: Optional precomputed scale of Fourier coefficients with shape
            :code:`(..., height, width // 2 + 1)`.

    Returns:
        y: Realization of the Gaussian process.
    """
    rfft2 = pack_rfft2(z) * _get_rfft2_scale(cov_rfft2, cov, rfft2_scale, z.shape[-1])
    return dispatch[rfft2].fft.irfft2(rfft2, z.shape[-2:]) + loc


def transform_rfft2(y: ArrayOrTensor, loc: ArrayOrTensor, *,
                    cov_rfft2: OptionalArrayOrTensor = None, cov: OptionalArrayOrTensor = None,
                    rfft2_scale: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Transform a Gaussian process realization to white noise in the Fourier domain.

    Args:
        y: Realization of the Gaussian process.
        loc: Mean of the Gaussian process with shape :code:`(..., height, width)`.
        cov_rfft2: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., height, width // 2 + 1)`.
        cov: Covariance between the first grid point and the remainder of the grid with shape
            :code:`(..., height, width)`.
        rfft2_scale: Optional precomputed scale of Fourier coefficients with shape
            :code:`(..., height, width // 2 + 1)`.

    Returns:
        z: Unpacked matrices with shape :code:`(..., height, width)`. See :func:`unpack_rfft2` for
            details.
    """
    rfft2_scale = _get_rfft2_scale(cov_rfft2, cov, rfft2_scale, y.shape[-1])
    return unpack_rfft2(dispatch[y].fft.rfft2(y - loc) / rfft2_scale, y.shape)


def evaluate_rfft2_log_abs_det_jacobian(width: int, *, cov_rfft2: OptionalArrayOrTensor = None,
                                        cov: OptionalArrayOrTensor = None,
                                        rfft2_scale: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Evaluate the log absolute determinant of the Jacobian associated with :func:`transform_rfft2`.

    Args:
        width: Number of columns of the signal (cannot be inferred from the Fourier coefficients).
        cov_rfft2: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., height, width // 2 + 1)`.
        cov: Covariance between the first grid point and the remainder of the grid with shape
            :code:`(..., height, width)`.
        rfft2_scale: Optional precomputed scale of Fourier coefficients with shape
            :code:`(..., height, width // 2 + 1)`.

    Returns:
        log_abs_det_jacobian: Log absolute determinant of the Jacobian.
    """
    rfft2_scale = _get_rfft2_scale(cov_rfft2, cov, rfft2_scale, width)
    height = rfft2_scale.shape[-2]
    assert rfft2_scale.shape[-1] == width // 2 + 1
    ncomplex_horizontal = (width - 1) // 2
    ncomplex_vertical = (height - 1) // 2
    parts = [
        # Real part of the first-column RFFT.
        - dispatch.log(rfft2_scale[..., :height // 2 + 1, 0]).sum(axis=-1),
        # Imaginary part of the first-column RFFT.
        - dispatch.log(rfft2_scale[..., 1:ncomplex_vertical + 1, 0]).sum(axis=-1),
        # Complex coefficients in subsequent columns.
        - 2 * dispatch.log(rfft2_scale[..., 1:1 + ncomplex_horizontal]).sum(axis=(-2, -1))
    ]
    # Account for Nyqvist frequencies in the last column if the number of columns is even.
    if width % 2 == 0:
        parts.extend([
            # Real part of the last-column RFFT.
            - dispatch.log(rfft2_scale[..., :height // 2 + 1, width // 2]).sum(axis=-1),
            # Imaginary part of the last-column RFFT.
            - dispatch.log(rfft2_scale[..., 1:ncomplex_vertical + 1, width // 2]).sum(axis=-1),
        ])
    size = width * height
    nterms = (size - 1) // 2
    if height % 2 == 0 and width % 2 == 0:
        nterms -= 1
    return sum(parts) - log2 * nterms + size * math.log(size) / 2


def evaluate_log_prob_rfft2(y: ArrayOrTensor, loc: ArrayOrTensor, *,
                            cov_rfft2: OptionalArrayOrTensor = None,
                            cov: OptionalArrayOrTensor = None,
                            rfft2_scale: OptionalArrayOrTensor = None) -> ArrayOrTensor:
    """
    Evaluate the log probability of a two-dimensional Gaussian process realization in Fourier space.

    Args:
        y: Realization of a Gaussian process with shape :code:`(..., height, width)`, where
            :code:`...` is the batch shape, :code:`height` is the number of rows, and :code:`width`
            is the number of columns.
        loc: Mean of the Gaussian process with shape :code:`(..., height, width)`.
        cov_rfft2: Precomputed real fast Fourier transform of the kernel with shape
            :code:`(..., height, width // 2 + 1)`.
        cov: Covariance between the first grid point and the remainder of the grid with shape
            :code:`(..., height, width)`.
        rfft2_scale: Optional precomputed scale of Fourier coefficients with shape
            :code:`(..., height, width // 2 + 1)`.

    Returns:
        log_prob: Log probability of the Gaussian process realization with shape :code:`(...)`.
    """
    rfft2_scale = _get_rfft2_scale(cov_rfft2, cov, rfft2_scale, y.shape[-1])
    rfft2 = transform_rfft2(y, loc, rfft2_scale=rfft2_scale)
    return log_prob_stdnorm(rfft2).sum(axis=(-2, -1)) \
        + evaluate_rfft2_log_abs_det_jacobian(y.shape[-1], rfft2_scale=rfft2_scale)



from gptools import util
from matplotlib import pyplot as plt
import numpy as np
import pytest
import re
import time
import torch as th
from typing import Tuple


@pytest.mark.parametrize("shape", [(3,), (4, 5)])
@pytest.mark.parametrize("ravel", [False, True])
def test_coord_grid(shape: Tuple[int], ravel: bool) -> None:
    xs = [np.arange(p) for p in shape]
    coords = util.coordgrid(*xs, ravel=ravel)
    if ravel:
        assert coords.shape == (np.prod(shape), len(shape),)
    else:
        assert coords.shape == shape + (len(shape),)


def test_timer(capsys: pytest.CaptureFixture):
    with util.Timer("timer test") as timer:
        time.sleep(0.1)
        assert 0.09 < timer.duration < 0.11
        time.sleep(0.05)
    time.sleep(0.1)
    assert 0.14 < timer.duration < 0.17
    out, _ = capsys.readouterr()
    assert re.match(r"timer test in 0\.1[56]", out)


def test_timer_errors():
    with util.Timer() as timer:
        time.sleep(1e-6)
    assert timer.duration > 0
    with pytest.raises(RuntimeError, match="timers can only"), timer:
        pass

    timer = util.Timer()
    with pytest.raises(RuntimeError, match="timer has not yet"):
        timer.duration


def test_dispatch():
    dispatch = util.ArrayOrTensorDispatch()

    tensor = th.empty(7)
    assert dispatch[tensor] is th
    assert dispatch[tensor, 0.1] is th
    assert isinstance(dispatch.add(tensor, tensor), th.Tensor)

    array = np.empty(3)
    assert dispatch[array] is np
    assert dispatch[array, 0.2] is np
    assert isinstance(dispatch.add(array, array), np.ndarray)

    with pytest.raises(ValueError):
        dispatch[tensor, array]
    with pytest.raises(ValueError):
        dispatch.add(array, tensor)


@pytest.mark.parametrize("float_, complex_", [
    (np.float32, np.complex64),
    (np.float64, np.complex128),
    (th.float32, th.complex64),
    (th.float64, th.complex128),
])
def test_get_complex_dtype(float_, complex_) -> None:
    dispatch = util.ArrayOrTensorDispatch()
    x = (th if float_.__module__ == "torch" else np).ones(3, dtype=float_)
    assert dispatch.get_complex_dtype(x) == complex_


def test_mutually_exclusive_kwargs() -> None:
    @util.mutually_exclusive_kwargs("a", ("b", "c"))
    def _func(a=None, b=None, c=None, given=None) -> None:
        return given

    assert _func(a=3) == "a"
    assert _func(b=2, c=1) == ("b", "c")
    with pytest.raises(ValueError):
        _func()
    with pytest.raises(ValueError):
        _func(a=3, b=2)
    with pytest.raises(ValueError):
        _func(b=2)
    with pytest.raises(ValueError):
        _func(a=3, b=2, c=1)


def test_encode_one_hot() -> None:
    x = np.random.randint(5, size=1000)
    y = util.encode_one_hot(x)
    assert y.shape == (x.shape[0], x.max() + 1)
    np.testing.assert_array_equal(x, np.argmax(y, axis=1))


@pytest.mark.parametrize("location", ["top", "right"])
def test_match_colorbar(location: str) -> None:
    fig, ax = plt.subplots()
    im = ax.imshow(np.random.normal(0, 1, (5, 5)))
    cb = fig.colorbar(im, ax=ax, location=location)
    util.match_colorbar(cb)



from .fft1 import evaluate_log_prob_rfft, evaluate_rfft_log_abs_det_jacobian, expand_rfft, \
    transform_irfft, transform_rfft
from .fft2 import evaluate_log_prob_rfft2, evaluate_rfft2_log_abs_det_jacobian, transform_irfft2, \
    transform_rfft2
from .util import log_prob_stdnorm


__all__ = [
    "evaluate_log_prob_rfft",
    "evaluate_log_prob_rfft2",
    "evaluate_rfft_log_abs_det_jacobian",
    "evaluate_rfft2_log_abs_det_jacobian",
    "expand_rfft",
    "log_prob_stdnorm",
    "transform_irfft",
    "transform_irfft2",
    "transform_rfft",
    "transform_rfft2",
]



from functools import partial
from gptools.util import Timer
from gptools.util.timeout import call_with_timeout
import pytest
import time


def target(wait: float, fail: bool) -> float:
    if fail:
        raise NotImplementedError("that didn't work")
    time.sleep(wait)
    return wait


@pytest.mark.parametrize("timeout, wait, fail", [
    (1, 0.5, False),
    (1, 0.5, True),
    (0.5, 1, False),
    (0.5, 1, True),
])
def test_timeout(timeout: float, wait: float, fail: bool) -> None:
    tol = 0.3  # Need some tolerance because the child takes some time to start.
    p = partial(call_with_timeout, timeout, target, wait, fail=fail)
    if fail:
        with Timer() as timer, pytest.raises(RuntimeError):
            p()
        assert timer.duration < tol
        return
    if timeout < wait:
        with Timer() as timer, pytest.raises(TimeoutError):
            p()
        assert timer.duration > timeout
        return
    with Timer() as timer:
        assert p() == wait
    assert timer.duration < wait + tol


def test_timeout_negative_timeout():
    with pytest.raises(ValueError):
        call_with_timeout(-1, None)


def test_timeout_not_a_callable():
    with pytest.raises(TypeError):
        call_with_timeout(3, None)



from gptools.util import coordgrid, kernels
from gptools.util.testing import KernelConfiguration
import itertools as it
import numpy as np
import pytest
from scipy.spatial.distance import cdist
import torch as th
from typing import Tuple


@pytest.mark.parametrize("shape", [(7,), (2, 3)])
def test_kernel(kernel_configuration: KernelConfiguration, shape: tuple, use_torch: bool) -> None:
    kernel = kernel_configuration()
    X = kernel_configuration.sample_locations(shape)
    cov = (kernel + kernels.DiagonalKernel(1e-3, kernel.period)).evaluate(X)
    # Check the shape and that the kernel is positive definite if there is nugget variance.
    *batch_shape, n = shape
    assert cov.shape == tuple(batch_shape) + (n, n)
    np.linalg.cholesky(cov)


@pytest.mark.parametrize("p", [1, 3, 5])
def test_evaluate_squared_distance(p: int, use_torch: bool) -> None:
    shape = (57, p)
    X = th.randn(shape) if use_torch else np.random.normal(0, 1, shape)
    np.testing.assert_allclose(kernels.evaluate_squared_distance(X), cdist(X, X) ** 2)


def test_periodic(kernel_configuration: KernelConfiguration):
    kernel = kernel_configuration()
    if not kernel.is_periodic:
        # Ensure the rfft cannot be evaluated and skip the rest.
        with pytest.raises(NotImplementedError):
            kernel.evaluate_rfft(tuple(range(13, 13 + len(kernel_configuration.dims))))
        return

    # Sample some points from the domain.
    X = kernel_configuration.sample_locations((13,))
    _, dim = X.shape
    cov = kernel.evaluate(X)
    # Ensure that translating by an integer period doesn't mess with the covariance.
    for delta in it.product(*([-1, 1, 2] for _ in range(dim))):
        Y = X + delta * np.asarray(kernel.period)
        other = kernel.evaluate(X[..., :, None, :], Y[..., None, :, :])
        np.testing.assert_allclose(cov, other, atol=1e-12)

    # For one and two dimensions, ensure that the Fourier transform has the correct structure.
    if dim > 2:
        return

    # We consider different shapes here to make sure we don't get unexpected behavior with edge
    # cases.
    for shape in it.product(*((5, 8) for _ in range(dim))):
        xs = kernel_configuration.coordgrid(shape)
        assert xs.shape == (np.prod(shape), len(shape))

        # Evaluate the covariance with the origin, take the Fourier transform, and check that there
        # is no imaginary part.
        cov = kernel.evaluate(xs)[0].reshape(shape)
        fftcov = np.fft.rfft(cov) if dim == 1 else np.fft.rfft2(cov)
        np.testing.assert_allclose(fftcov.imag, 0, atol=1e-9)

        # We may want to check that the numerical and theoretic FFT match, but this requires a more
        # "proper" implementation of the periodic kernels involving infinite sums (see
        # https://github.com/tillahoffmann/gptools/issues/59 for details). For now, let's verify we
        # have a positive-definite kernel.
        np.testing.assert_array_less(-1e-12, kernel.evaluate_rfft(shape))


def test_kernel_composition():
    a = kernels.ExpQuadKernel(2, 0.5)
    b = kernels.DiagonalKernel()
    kernel = a + 1 + b
    x = np.random.normal(0, 1, (100, 2))
    np.testing.assert_allclose(kernel.evaluate(x), a.evaluate(x) + 1 + b.evaluate(x))


def test_kernel_composition_period():
    with pytest.raises(ValueError):
        kernels.DiagonalKernel(period=1) + kernels.DiagonalKernel()
    with pytest.raises(ValueError):
        kernels.DiagonalKernel(period=1) + kernels.DiagonalKernel(period=2)


def test_diagonal_kernel():
    kernel = kernels.DiagonalKernel()
    x = np.random.normal(0, 1, (10, 2))
    np.testing.assert_allclose(kernel.evaluate(x), np.eye(10))
    with pytest.raises(ValueError):
        kernel.evaluate(x, x)


@pytest.mark.parametrize("size", [500, (501,), (500, 502), (500, 501), (501, 500), (501, 503)])
def test_periodic_exp_quad_rfft(size: int) -> None:
    shape = [size] if isinstance(size, int) else size
    *head, tail = shape
    ndim = len(shape)
    # Use a large number of terms to evaluate the kernel.
    if ndim == 1:
        kernel = kernels.ExpQuadKernel(1.2, 0.1, 3)
    elif ndim == 2:
        kernel = kernels.ExpQuadKernel(0.9, np.asarray([0.2, 0.3]), np.asarray([2.1, 2.3]))
    else:
        raise ValueError
    xs = coordgrid(*[np.linspace(0, period, n, endpoint=False) for n, period in
                     zip(shape, kernel.period * np.ones(ndim))])
    cov = kernel.evaluate(0, xs).reshape(size)
    if ndim == 1:
        rfft = np.fft.rfft(cov)
    elif ndim == 2:
        rfft = np.fft.rfft2(cov)
    else:
        raise ValueError
    assert rfft.shape == (*head, tail // 2 + 1,)
    np.testing.assert_allclose(rfft.imag, 0, atol=1e-9)

    direct_rfft = kernel.evaluate_rfft(size)
    poly = np.polynomial.Polynomial.fit(rfft.ravel(), direct_rfft.ravel(), 1).convert()
    bias, slope = poly.coef
    assert abs(bias) < 1e-2
    assert abs(slope - 1) < 1e-2


def test_matern_invalid_dof() -> None:
    with pytest.raises(ValueError):
        kernels.MaternKernel(1, 1, 1)


@pytest.mark.parametrize("dof", [3 / 2, 5 / 2])
@pytest.mark.parametrize("size", [500, (501,), (500, 502), (500, 501), (501, 500), (501, 503)])
def test_matern_approximate_rfft(dof: float, size: Tuple[int]) -> None:
    sigma = 1.2
    shape = [size] if isinstance(size, int) else size
    ndim = len(shape)
    period = np.asarray([2.1, 1.7])[:ndim]
    length_scale = np.asarray([0.01, 0.02])[:ndim]
    kernel = kernels.MaternKernel(dof, sigma, length_scale)
    periodic_kernel = kernels.MaternKernel(dof, sigma, length_scale, period)
    xs = coordgrid(*(np.linspace(0, p, n, endpoint=False) for n, p in zip(shape, period)))
    xs = np.minimum(xs, period - xs)
    cov = kernel.evaluate(np.zeros(ndim), xs).reshape(shape)
    if ndim == 1:
        rfft = np.fft.rfft(cov)
    elif ndim == 2:
        rfft = np.fft.rfft2(cov)
    else:
        raise ValueError(ndim)
    np.testing.assert_allclose(rfft.imag, 0, atol=1e-9)
    rfft = rfft.real
    direct_rfft = periodic_kernel.evaluate_rfft(size)
    poly = np.polynomial.Polynomial.fit(rfft.ravel(), direct_rfft.ravel(), 1).convert()
    bias, slope = poly.coef
    assert abs(bias) < 1e-2
    assert abs(slope - 1) < 1e-2



from setuptools import find_namespace_packages, setup

with open("README.rst") as fp:
    long_description = fp.read()

setup(
    name="gptools-util",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    packages=find_namespace_packages(),
    version="0.1.1",
    python_requires=">=3.8",
    install_requires=[
        "numpy",
    ],
    extras_require={
        "docs": [
            "myst-nb",
            "sphinx",
            "sphinx-multiproject",
            "sphinx_rtd_theme",
        ],
        "tests": [
            "doit-interface",
            "flake8",
            "jupyter",
            "matplotlib",
            "networkx",
            "pytest",
            "pytest-cov",
            "scipy",
            "torch",
            "twine",
        ],
    }
)



from myst_parser.config.main import MdParserConfig
from myst_nb.core.config import NbParserConfig
from myst_nb.core.read import create_nb_reader, NbReader
from nbconvert.preprocessors import ExecutePreprocessor
import numpy as np
import os
import pathlib
import pytest
from typing import Any, Callable, Type
from unittest import mock
from .kernels import ExpQuadKernel, Kernel, MaternKernel
from . import coordgrid, ArrayOrTensor


class KernelConfiguration:
    """
    Kernel configuration for testing.

    Args:
        dims: Sequence of dimension domains. :code:`None` indicates an unbounded domain.
        kernel_cls: Class of the kernel to create.
        **kwargs: Keyword arguments passed to the kernel.
    """
    def __init__(self, dims: tuple, kernel_cls: Type[Kernel], **kwargs) -> None:
        if all(dims):
            dims = np.asarray(dims)
        self.dims = dims
        self.kernel_cls = kernel_cls
        self.kwargs = kwargs
        if all(self.dims):
            self.kwargs.setdefault("period", self.dims)

    def __call__(self) -> Kernel:
        return self.kernel_cls(**self.kwargs)

    def sample_locations(self, size: tuple = None) -> ArrayOrTensor:
        """
        Sample locations consistent with the domain on which to apply the kernel.
        """
        locations = []
        for dim in self.dims:
            if dim is None:
                locations.append(np.random.normal(0, 1, size))
            else:
                domain = dim
                locations.append(np.random.uniform(0, domain, size))
        locations = np.asarray(locations)
        return np.moveaxis(locations, 0, -1)

    def coordgrid(self, shape) -> ArrayOrTensor:
        """
        Create a coordinate grid on which to evaluate the kernel.
        """
        lins = []
        for dim, n in zip(self.dims, shape):
            if dim is None:  # pragma: no cover
                raise ValueError("cannot create coordinate grid for unbounded support")
            lins.append(np.linspace(0, dim, n, endpoint=False))
        return coordgrid(*lins)


_kernel_configurations = [
    KernelConfiguration([None], ExpQuadKernel, sigma=1.3, length_scale=0.2),
    KernelConfiguration([None, None, None], ExpQuadKernel, sigma=4, length_scale=0.2),
    KernelConfiguration([2, 3], ExpQuadKernel, sigma=1.7, length_scale=0.3),
    KernelConfiguration([1.5], ExpQuadKernel, sigma=1.5, length_scale=0.1),
    KernelConfiguration([2, 3, 4], ExpQuadKernel, sigma=2.1,
                        length_scale=np.asarray([0.1, 0.15, 0.2])),
    KernelConfiguration([None], MaternKernel, dof=3 / 2, sigma=1.3, length_scale=0.2),
    KernelConfiguration([None, None, None], MaternKernel, dof=3 / 2, sigma=4, length_scale=0.2),
    KernelConfiguration([None], MaternKernel, dof=5 / 2, sigma=1.3, length_scale=0.2),
    KernelConfiguration([None, None, None], MaternKernel, dof=5 / 2, sigma=4, length_scale=0.2),
]


@pytest.fixture(params=_kernel_configurations)
def kernel_configuration(request: pytest.FixtureRequest) -> KernelConfiguration:
    return request.param


EXCLUDED_DIRECTORIES = {".ipynb_checkpoints"}


def run_myst_notebook(path: pathlib.Path) -> Any:  # pragma: no cover
    """
    Run a myst example notebook.
    """
    path = pathlib.Path(path)
    timeout = 60 if "CI" in os.environ else None
    md_config = MdParserConfig()
    nb_config = NbParserConfig()
    with open(path) as fp:
        content = fp.read()
    reader: NbReader = create_nb_reader(str(path), md_config, nb_config, content)
    notebook = reader.read(content)
    with mock.patch.dict(os.environ, CI="true"):
        preprocessor = ExecutePreprocessor(timeout=timeout)
        return preprocessor.preprocess(notebook, {"metadata": {"path": path.parent}})


def parameterized_notebook_tests(file: str, rel: str) -> Callable:  # pragma: no cover
    """
    Discover notebooks starting at a root directory and return a parameterized test to execute all.
    """
    root = (pathlib.Path(file).parent / rel).resolve()
    notebooks = [str(notebook) for notebook in root.glob("**/*.md") if not
                 any(x in notebook.parts for x in EXCLUDED_DIRECTORIES)]

    @pytest.mark.parametrize("notebook", notebooks)
    def _wrapped(notebook: str) -> None:
        run_myst_notebook(notebook)
    return _wrapped
